﻿// This code has been provided by Andrew Conrad from Microsoft
// See http://blogs.msdn.com/aconrad/archive/2007/09/07/science-project.aspx

using System;
using System.Collections.Generic;
using System.Data;
using System.Reflection;

namespace LinqInAction.Chapter14
{
  public class ObjectShredder<T>
  {
    private FieldInfo[] _fi;
    private PropertyInfo[] _pi;
    private Dictionary<string, int> _ordinalMap;
    private Type _type;

    public ObjectShredder()
    {
      _type = typeof(T);
      _fi = _type.GetFields();
      _pi = _type.GetProperties();
      _ordinalMap = new Dictionary<string, int>();
    }

    public DataTable Shred(IEnumerable<T> source, DataTable table, LoadOption? options)
    {
      if (typeof(T).IsPrimitive)
      {
        return ShredPrimitive(source, table, options);
      }

      if (table == null)
      {
        table = new DataTable(typeof(T).Name);
      }

      // now see if need to extend datatable base on the type T + build ordinal map
      table = ExtendTable(table, typeof(T));

      table.BeginLoadData();
      using (IEnumerator<T> e = source.GetEnumerator())
      {
        while (e.MoveNext())
        {
          if (options != null)
          {
            table.LoadDataRow(ShredObject(table, e.Current), (LoadOption)options);
          }
          else
          {
            table.LoadDataRow(ShredObject(table, e.Current), true);
          }
        }
      }
      table.EndLoadData();
      return table;
    }

    public DataTable ShredPrimitive(IEnumerable<T> source, DataTable table, LoadOption? options)
    {
      if (table == null)
      {
        table = new DataTable(typeof(T).Name);
      }

      if (!table.Columns.Contains("Value"))
      {
        table.Columns.Add("Value", typeof(T));
      }

      table.BeginLoadData();
      using (IEnumerator<T> e = source.GetEnumerator())
      {
        Object[] values = new Object[table.Columns.Count];
        while (e.MoveNext())
        {
          values[table.Columns["Value"].Ordinal] = e.Current;

          if (options != null)
          {
            table.LoadDataRow(values, (LoadOption)options);
          }
          else
          {
            table.LoadDataRow(values, true);
          }
        }
      }
      table.EndLoadData();
      return table;
    }

    public DataTable ExtendTable(DataTable table, Type type)
    {
      // value is type derived from T, may need to extend table.
      foreach (FieldInfo f in type.GetFields())
      {
        if (!_ordinalMap.ContainsKey(f.Name))
        {
          DataColumn dc = table.Columns.Contains(f.Name) ? table.Columns[f.Name]
                        : table.Columns.Add(f.Name, f.FieldType);
          _ordinalMap.Add(f.Name, dc.Ordinal);
        }
      }
      foreach (PropertyInfo p in type.GetProperties())
      {
        if (!_ordinalMap.ContainsKey(p.Name))
        {
          DataColumn dc = table.Columns.Contains(p.Name) ? table.Columns[p.Name]
                        : table.Columns.Add(p.Name, p.PropertyType);
          _ordinalMap.Add(p.Name, dc.Ordinal);
        }
      }
      return table;
    }

    public object[] ShredObject(DataTable table, T instance)
    {

      FieldInfo[] fi = _fi;
      PropertyInfo[] pi = _pi;

      if (instance.GetType() != typeof(T))
      {
        ExtendTable(table, instance.GetType());
        fi = instance.GetType().GetFields();
        pi = instance.GetType().GetProperties();
      }

      Object[] values = new object[table.Columns.Count];
      foreach (FieldInfo f in fi)
      {
        values[_ordinalMap[f.Name]] = f.GetValue(instance);
      }

      foreach (PropertyInfo p in pi)
      {
        values[_ordinalMap[p.Name]] = p.GetValue(instance, null);
      }
      return values;
    }
  }
}